import io
import logging
from pathlib import Path
from typing import List, Optional

import easyocr
from PIL import Image
from pypdf import PdfReader
from pypdf.errors import PdfReadError

logger = logging.getLogger(__name__)


class TextExtractor:
    """Handles text extraction and processing."""

    def __init__(self):
        """Initialize the text extractor."""
        self.reader = easyocr.Reader(["en"])

    def extract_text_from_pdf(self, pdf_path: str) -> Optional[str]:
        """Extract raw text content from a PDF file."""
        if not Path(pdf_path).exists():
            raise FileNotFoundError(f"PDF file not found: {pdf_path}")

        try:
            text_content = []
            with open(pdf_path, "rb") as file:
                reader = PdfReader(file)
                if len(reader.pages) == 0:
                    return ""

                for page in reader.pages:
                    text = page.extract_text()
                    if text.strip():
                        text_content.append(text)

            return "\n".join(text_content)
        except PdfReadError as e:
            logger.error(f"Failed to read PDF: {str(e)}")
            return ""
        except Exception as e:
            logger.error("Failed to extract text from PDF: %s", str(e))
            raise ValueError(f"Failed to extract text from PDF: {str(e)}") from e

    def extract_text_from_image(self, image: Image.Image) -> str:
        """Extract text from an image using OCR.

        Args:
            image_path: Path to the image file

        Returns:
            str: Extracted text content
        """
        try:
            # Convert PIL Image to bytes
            img_byte_arr = io.BytesIO()
            image.save(img_byte_arr, format="PNG")
            img_byte_arr = img_byte_arr.getvalue()

            # Perform OCR
            results = self.reader.readtext(img_byte_arr)

            # Extract text from results
            text = " ".join([result[1] for result in results])

            return text.strip()

        except Exception as e:
            logger.error(f"Failed to extract text from image: {str(e)}")
            raise ValueError(f"Failed to extract text from image: {str(e)}") from e

    def extract_text_from_images(self, images: List[Image.Image]) -> str:
        """Extract text from multiple images."""
        texts = []
        for image in images:
            text = self.extract_text_from_image(image)
            if text:
                texts.append(text)

        return "\n".join(texts)



import pytest

from qz.text_extractor import TextExtractor


def test_text_extractor_initialization():
    extractor = TextExtractor()
    assert isinstance(extractor, TextExtractor)


def test_extract_text_from_pdf(sample_pdf_path):
    extractor = TextExtractor()
    text = extractor.extract_text_from_pdf(sample_pdf_path)
    assert isinstance(text, str)


def test_nonexistent_pdf():
    extractor = TextExtractor()
    with pytest.raises(FileNotFoundError):
        extractor.extract_text_from_pdf("nonexistent.pdf")



"""Tests for the QuestionGenerator class."""

from unittest.mock import Mock, patch

import numpy as np
import pytest
from spacy.language import Language

from qz.question_generator import QuestionGenerator


@pytest.fixture(name="spacy_nlp")
def fixture_spacy_nlp():
    """Create a mock spacy NLP object."""
    nlp = Mock(spec=Language)
    doc_mock = Mock()
    doc_mock.sents = [
        Mock(text="Test sentence 1", root="test"),
        Mock(text="Test sentence 2", root="test"),
    ]
    nlp.return_value = doc_mock
    return nlp


@pytest.fixture(name="test_transformer")
def fixture_sentence_transformer():
    """Create a mock sentence transformer."""
    mock = Mock()
    mock.encode.return_value = np.array([[0.1, 0.2], [0.3, 0.4]])
    return mock


def test_initialization():
    """Test QuestionGenerator initialization."""
    with patch("spacy.load") as mock_load:
        mock_load.return_value = Mock(spec=Language)
        generator = QuestionGenerator()
        assert isinstance(generator, QuestionGenerator)
        assert generator.qa_model is not None
        assert generator.qa_tokenizer is not None
        assert generator.device is not None


def test_generate_question_from_text(spacy_nlp):
    """Test question generation from text."""
    with (
        patch("spacy.load", return_value=spacy_nlp),
        patch("transformers.AutoTokenizer.from_pretrained"),
        patch("transformers.AutoModelForQuestionAnswering.from_pretrained"),
        patch("transformers.BlipProcessor.from_pretrained"),
        patch("transformers.BlipForConditionalGeneration.from_pretrained"),
        patch("sentence_transformers.SentenceTransformer"),
    ):
        generator = QuestionGenerator()
        result = generator.generate_question_from_text("Test context")

        assert isinstance(result, dict)
        assert "text" in result
        assert "context" in result
        assert "answer" in result
        assert result["context"] == "Test context"


def test_generate_distractors(spacy_nlp, test_transformer):
    """Test distractor generation."""
    generator = QuestionGenerator(nlp=spacy_nlp, sentence_model=test_transformer)
    distractors = generator.generate_distractors(
        correct_answer="test answer", context="Test context with some content", num_distractors=3
    )
    assert isinstance(distractors, list)
    assert len(distractors) == 3


def test_create_quiz_question():
    """Test complete quiz question creation."""
    with (
        patch.object(QuestionGenerator, "generate_question_from_text") as mock_gen_question,
        patch.object(QuestionGenerator, "generate_distractors") as mock_gen_distractors,
    ):
        mock_gen_question.return_value = {
            "text": "Test question?",
            "context": "Test context",
            "answer": "correct answer",
        }
        mock_gen_distractors.return_value = ["wrong1", "wrong2", "wrong3"]

        generator = QuestionGenerator()
        result = generator.create_quiz_question("Test context")

        assert isinstance(result, dict)
        assert "text" in result
        assert "options" in result
        assert "correct_answer" in result
        assert len(result["options"]) == 4
        assert result["options"][result["correct_answer"]] == "correct answer"
        assert "difficulty" in result
        assert "category" in result



import logging

import ollama
from ollama._types import ResponseError

logger = logging.getLogger(__name__)


class FactChecker:
    """Handles fact checking of text."""

    def __init__(self, model_name: str = "deepseek-coder"):
        """Initialize the fact checker.

        Args:
            model_name: Name of the Ollama model to use
        """
        self.model_name = model_name

    def _safe_generate(self, prompt: str) -> str:
        """Safely generate response from Ollama."""
        try:
            # Add system context for deepseek-coder
            system_prompt = "You are a fact-checking assistant. Respond with 'yes' or 'no' only."
            response = ollama.generate(model=self.model_name, prompt=prompt, system=system_prompt)
            # Extract response content from the dictionary
            return response.get("response", "")
        except ResponseError as e:
            logger.error(f"Ollama error: {e}")
            return ""
        except Exception as e:
            logger.error(f"Unexpected error calling Ollama: {e}")
            return ""

    def validate_question(self, question: str, context: str) -> bool:
        """Validate if the question is relevant to the context.

        Args:
            question: The generated question
            context: The context from which the question was generated

        Returns:
            bool: True if the question is valid, False otherwise
        """
        question_text = str(question).lower()
        context_text = str(context).lower()
        prompt = f"Is the following question relevant to the context? Question: {question_text} Context: {context_text}"
        response = self._safe_generate(prompt)

        # Check if the response indicates validity
        return "yes" in response.lower()

    def fact_check(self, question: str, answer: str, context: str) -> bool:
        """Check if the answer to the question is factually correct.

        Args:
            question: The generated question
            answer: The generated answer
            context: The context from which the question and answer were generated

        Returns:
            bool: True if the answer is factually correct, False otherwise
        """
        question_text = str(question).lower()
        answer_text = str(answer).lower()
        context_text = str(context).lower()
        prompt = f"Is the following answer correct? Question: {question_text} Answer: {answer_text} Context: {context_text}"
        response = self._safe_generate(prompt)

        # Check if the response indicates correctness
        return "yes" in response.lower()



import re
from typing import List


class TextPreprocessor:
    """Handles text preprocessing and chunking."""

    def preprocess_text(self, text: str) -> str:
        """Normalize and clean text content.

        Args:
            text: Raw text content

        Returns:
            Cleaned and normalized text
        """
        # Remove extra whitespace
        text = re.sub(r"\s+", " ", text.strip())

        # Remove special characters but keep basic punctuation
        text = re.sub(r"[^a-zA-Z0-9\s.,!?]", "", text)

        # Split into sentences based on any punctuation that ends a sentence,
        # keeping the punctuation with the sentence
        sentences = re.findall(r"[^.!?]+[.!?]", text)

        # Remove extra whitespace from each sentence and add newline
        processed_sentences = [s.strip() + "\n" for s in sentences]

        return "".join(processed_sentences).strip()

    def _process_single_sentence(self, sentence: str, chunk_size: int) -> List[str]:
        """Process a single sentence that exceeds chunk size."""
        chunks = []
        words = sentence.split()
        current_chunk = []
        current_length = 0

        for word in words:
            word_len = len(word) + 1  # +1 for space
            if current_length + word_len > chunk_size:
                if current_chunk:
                    chunks.append(" ".join(current_chunk))
                current_chunk = [word]
                current_length = word_len
            else:
                current_chunk.append(word)
                current_length += word_len

        if current_chunk:
            chunks.append(" ".join(current_chunk))
        return chunks

    def split_into_chunks(self, text: str, chunk_size: int = 500) -> List[str]:
        """Split text into chunks of approximately equal size."""
        if not text or not text.strip():
            return []

        # Split into sentences first
        sentences = re.split(r"(?<=[.!?])\s+", text)
        chunks = []
        current_chunk = []
        current_length = 0

        for sentence in sentences:
            sentence_length = len(sentence)

            # If single sentence is longer than chunk_size, split it
            if sentence_length > chunk_size:
                if current_chunk:
                    chunks.append(" ".join(current_chunk))
                    current_chunk = []
                    current_length = 0

                chunks.extend(self._process_single_sentence(sentence, chunk_size))
                continue

            # Regular sentence handling
            if current_length + sentence_length + 1 > chunk_size:
                if current_chunk:
                    chunks.append(" ".join(current_chunk))
                current_chunk = [sentence]
                current_length = sentence_length
            else:
                current_chunk.append(sentence)
                current_length += sentence_length + 1

        # Add the last chunk if it exists
        if current_chunk:
            chunks.append(" ".join(current_chunk))

        return chunks



from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Dict, List

from qz.fact_checker import FactChecker
from qz.question_generator import QuestionGenerator
from qz.text_extractor import TextExtractor
from qz.text_preprocessor import TextPreprocessor


class PDFExtractor:
    """Handles PDF text extraction."""

    def __init__(self):
        """Initialize the PDF extractor with required components."""
        self.text_extractor = TextExtractor()
        self.text_preprocessor = TextPreprocessor()
        self.question_generator = QuestionGenerator()
        self.fact_checker = FactChecker()

    def generate_quiz(self, pdf_path: str, num_questions: int = 5) -> List[Dict[str, Any]]:
        """Generate a quiz from a PDF file.

        Args:
            pdf_path: Path to the PDF file
            num_questions: Number of questions to generate

        Returns:
            List of questions with their answers and options
        """
        # Extract text from PDF
        text = self.text_extractor.extract_text_from_pdf(pdf_path)
        if not text:
            return []

        # Preprocess text
        chunks = self.text_preprocessor.split_into_chunks(text)
        questions = []
        failed_chunks = []

        # Generate questions in parallel
        with ThreadPoolExecutor() as executor:
            futures = []
            for chunk in chunks[:num_questions]:
                futures.append(executor.submit(self.question_generator.create_quiz_question, chunk))

            for future in as_completed(futures):
                try:
                    question = future.result()
                    if question and self.fact_checker.validate_question(
                        question["text"], question["context"]
                    ):
                        questions.append(question)
                except Exception as e:
                    failed_chunks.append(str(e))
                    if len(failed_chunks) == len(chunks):
                        # If all chunks failed, something is seriously wrong
                        raise RuntimeError(
                            f"Failed to generate questions from any chunk. First error: {failed_chunks[0]}"
                        ) from e

        return questions[:num_questions]



from qz.quality import is_valid_question


def test_is_valid_question():
    # Valid case
    assert is_valid_question(
        "What is the capital of France?", ["Paris", "London", "Berlin"], "Paris"
    )

    # Invalid cases
    assert not is_valid_question("", ["A", "B"], "A")  # Empty question
    assert not is_valid_question("Test?", ["A", "B"], "C")  # Answer not in options
    assert not is_valid_question("Test?", ["A", "A"], "A")  # Duplicate options
    assert not is_valid_question("Test?", ["A"], "A")  # Too few options



import argparse
import logging

from qz.quality import is_valid_question
from qz.question_generator import QuestionGenerator
from qz.text_extractor import TextExtractor


def generate_questions(text: str, num_questions: int) -> list:
    generator = QuestionGenerator()
    questions = []
    attempts = 0
    max_attempts = num_questions * 2

    while len(questions) < num_questions and attempts < max_attempts:
        question_data = generator.create_quiz_question(text)
        if is_valid_question(
            question_data["text"], question_data["options"], question_data["correct_answer"]
        ):
            questions.append(question_data)
        attempts += 1

    return questions


def main():
    # Set up argument parser
    parser = argparse.ArgumentParser(description="Generate a quiz from a PDF file.")
    parser.add_argument("pdf_path", type=str, help="Path to the PDF file")
    parser.add_argument(
        "--num_questions",
        type=int,
        default=5,
        help="Number of questions to generate (default: 5)",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=4,
        help="Number of questions to process in parallel (default: 4)",
    )

    # Parse arguments
    args = parser.parse_args()

    try:
        # Extract text from PDF
        text_extractor = TextExtractor()
        text = text_extractor.extract_text_from_pdf(args.pdf_path)

        # Generate quiz questions
        quiz = generate_questions(text, args.num_questions)

        # Display the quiz
        for i, question in enumerate(quiz):
            print(f"Question {i + 1}: {question['text']}")
            print(f"Options: {question['options']}")
            print(f"Correct Answer: {question['options'][question['correct_answer']]}")
            print()
    except FileNotFoundError as e:
        logging.error("File not found: %s", str(e))
    except ValueError as e:
        logging.error("Invalid input: %s", str(e))
    except Exception as e:
        logging.error("Unexpected error: %s", str(e))
        raise


if __name__ == "__main__":
    main()



from unittest.mock import patch

from reportlab.pdfgen import canvas

from qz.main import main


def test_main_successful_execution(capsys, tmp_path):
    mock_quiz = [{"text": "Test question?", "options": ["correct", "wrong"], "correct_answer": 0}]

    # Create a valid PDF file
    test_pdf = tmp_path / "test.pdf"
    c = canvas.Canvas(str(test_pdf))
    c.drawString(100, 750, "This is a test PDF document.")
    c.save()

    with (
        patch("argparse.ArgumentParser.parse_args") as mock_args,
        patch("qz.text_extractor.TextExtractor.extract_text_from_pdf", return_value="Test content"),
        patch("qz.main.generate_questions", return_value=mock_quiz),
    ):
        # Configure arguments
        mock_args.return_value.pdf_path = str(test_pdf)
        mock_args.return_value.num_questions = 1
        mock_args.return_value.batch_size = 1

        # Run main
        main()

        # Verify output
        captured = capsys.readouterr()
        assert "Test question?" in captured.out
        assert "correct" in captured.out


def test_main_file_not_found():
    with (
        patch("argparse.ArgumentParser.parse_args") as mock_args,
        patch("qz.pdf_extractor.PDFExtractor") as mock_extractor,
        patch("logging.error") as mock_error,
    ):
        mock_args.return_value.pdf_path = "nonexistent.pdf"
        mock_extractor.return_value.generate_quiz.side_effect = FileNotFoundError()

        main()
        mock_error.assert_called_once()



from qz.text_preprocessor import TextPreprocessor


def test_preprocess_text():
    preprocessor = TextPreprocessor()
    text = "  Hello,   World!  This is a test. With multiple sentences?  "
    result = preprocessor.preprocess_text(text)
    assert "Hello, World!" in result
    assert "This is a test." in result
    assert "With multiple sentences?" in result
    assert "  " not in result  # No double spaces


def test_split_into_chunks():
    preprocessor = TextPreprocessor()
    text = "First sentence. Second sentence. Third sentence. Fourth sentence."
    chunks = preprocessor.split_into_chunks(text, chunk_size=30)
    assert len(chunks) > 1
    assert all(len(chunk) <= 30 for chunk in chunks)
    assert all(chunk.endswith(".") for chunk in chunks)


def test_empty_text():
    preprocessor = TextPreprocessor()
    assert preprocessor.preprocess_text("") == ""
    assert not preprocessor.split_into_chunks("")  # Simplified boolean check



from qz.fact_checker import FactChecker


def test_fact_checker_initialization():
    checker = FactChecker()
    assert isinstance(checker, FactChecker)
    assert checker.model_name == "deepseek-coder"


def test_validate_question(mock_ollama):
    checker = FactChecker()
    question = "What is testing?"
    context = "Testing is a process of evaluating software."
    result = checker.validate_question(question, context)
    assert isinstance(result, bool)
    mock_ollama.assert_called_once()


def test_fact_check(mock_ollama):
    checker = FactChecker()
    result = checker.fact_check(
        question="What is testing?",
        answer="A process of evaluating software",
        context="Testing is a process of evaluating software.",
    )
    assert isinstance(result, bool)
    mock_ollama.assert_called_once()


def test_ollama_error_handling():
    checker = FactChecker(model_name="nonexistent-model")
    # Should return False when Ollama fails
    assert not checker.validate_question("test?", "test.")
    assert not checker.fact_check("test?", "test", "test.")


def test_default_model_name():
    checker = FactChecker()
    assert checker.model_name == "deepseek-coder"



from setuptools import find_packages, setup

# Read requirements using context manager
with open("requirements.txt", encoding="utf-8") as f:
    requirements = [line.strip() for line in f if line.strip() and not line.startswith("#")]

setup(
    name="qz",
    version="0.1",
    packages=find_packages(),
    install_requires=requirements,
)



import pytest

from qz.pdf_extractor import PDFExtractor


def test_pdf_extractor_initialization():
    extractor = PDFExtractor()
    assert isinstance(extractor, PDFExtractor)
    assert extractor.text_extractor is not None
    assert extractor.text_preprocessor is not None
    assert extractor.question_generator is not None
    assert extractor.fact_checker is not None


def test_generate_quiz(sample_pdf_path):
    extractor = PDFExtractor()
    quiz = extractor.generate_quiz(sample_pdf_path, num_questions=2)
    assert isinstance(quiz, list)
    if quiz:  # If questions were generated
        assert isinstance(quiz[0], dict)
        assert "text" in quiz[0]
        assert "options" in quiz[0]
        assert "correct_answer" in quiz[0]


def test_invalid_pdf_path():
    extractor = PDFExtractor()
    with pytest.raises(FileNotFoundError):
        extractor.generate_quiz("nonexistent.pdf")



from functools import lru_cache
from typing import Any, Dict, List, Optional

import numpy as np
import spacy
import torch
from sentence_transformers import SentenceTransformer
from spacy.language import Language
from transformers import (
    AutoModelForQuestionAnswering,
    AutoTokenizer,
    BlipForConditionalGeneration,
    BlipProcessor,
)


class QuestionGenerator:  # pylint: disable=too-many-instance-attributes
    """Handles question generation from text and images."""

    def __init__(self, nlp: Optional[Language] = None, sentence_model=None):
        """Initialize the question generator with models."""
        # Initialize model configuration first
        self.qa_model_name = "distilbert-base-cased-distilled-squad"

        # Initialize tokenizer and model
        self.qa_tokenizer = self._load_qa_tokenizer()
        self.qa_model = self._load_qa_model()
        self.blip_processor = self._load_blip_processor()
        self.blip_model = self._load_blip_model()

        # Allow injection for testing
        if nlp is not None:
            self.nlp = nlp
        else:
            try:
                self.nlp = spacy.load(
                    "en_core_web_sm", exclude=["vectors", "textcat", "parser", "ner"]
                )
            except BaseException:  # Catch without binding to unused variable
                self.nlp = spacy.blank("en")
                self.nlp.add_pipe("sentencizer")

        self.sentence_model = sentence_model or SentenceTransformer("all-MiniLM-L6-v2")
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # Move models to appropriate device
        self.qa_model.to(self.device)
        self.blip_model.to(self.device)

    @lru_cache(maxsize=1)  # Cache model loading
    def _load_qa_tokenizer(self):
        return AutoTokenizer.from_pretrained(self.qa_model_name)

    @lru_cache(maxsize=1)  # Cache model loading
    def _load_qa_model(self):
        return AutoModelForQuestionAnswering.from_pretrained(self.qa_model_name)

    @lru_cache(maxsize=1)  # Cache model loading
    def _load_blip_processor(self):
        return BlipProcessor.from_pretrained("Salesforce/blip-image-captioning-base")

    @lru_cache(maxsize=1)  # Cache model loading
    def _load_blip_model(self):
        return BlipForConditionalGeneration.from_pretrained("Salesforce/blip-image-captioning-base")

    def generate_question_from_text(self, context: str) -> Dict[str, Any]:
        """Generate a question from a given text context."""
        doc = self.nlp(context)
        sentences = list(doc.sents)
        if not sentences:
            raise ValueError("No sentences found in text")

        sentence = np.random.choice(sentences)
        question = f"Which of the following is correct about {sentence.root}?"
        answer = str(sentence)

        return {"text": question, "context": context, "answer": answer}

    def generate_distractors(
        self, correct_answer: str, context: str, num_distractors: int = 3
    ) -> List[str]:
        """Generate distractor options using semantic similarity."""
        if not correct_answer or not context:
            return ["Option A", "Option B", "Option C"]  # Default fallback

        # Extract sentences from context
        try:
            doc = self.nlp(context)
            sentences = [sent.text.strip() for sent in doc.sents if sent.text.strip()]

            if not sentences:
                return ["Option A", "Option B", "Option C"]

            # Convert sentences to numpy array for encoding
            sentences_array = np.array(sentences)
            sentence_embeddings = self.sentence_model.encode(sentences_array.tolist())
            correct_embedding = self.sentence_model.encode([correct_answer])[0]

            # Calculate similarities
            similarities = np.dot(sentence_embeddings, correct_embedding)

            # Get most dissimilar sentences as distractors
            sorted_indices = np.argsort(similarities)
            distractors = []

            for idx in sorted_indices:
                if len(distractors) >= num_distractors:
                    break
                if sentences[idx] != correct_answer:
                    distractors.append(sentences[idx])

            # If we don't have enough distractors, add some default ones
            while len(distractors) < num_distractors:
                distractors.append(f"Option {len(distractors) + 1}")

            return distractors[:num_distractors]

        except Exception:
            return ["Option A", "Option B", "Option C"]

    def create_quiz_question(self, context: str) -> Dict[str, Any]:
        """Create a complete quiz question with distractors."""
        question_data = self.generate_question_from_text(context)
        distractors = self.generate_distractors(question_data["answer"], context)
        options = [question_data["answer"]] + distractors
        np.random.shuffle(options)

        correct_index = options.index(question_data["answer"])

        return {
            "text": question_data["text"],
            "options": options,
            "correct_answer": correct_index,
            "context": context,
            "difficulty": "medium",
            "category": "auto-generated",
        }



def is_valid_question(question: str, options: list, answer: str) -> bool:
    """Validate if a generated Q&A pair meets quality standards."""

    # Initial validation checks
    if not all((question, options, answer)):
        return False

    # Length and content checks
    if len(question.split()) <= 2 or len(question) < 10:
        return False

    # Answer validation
    if answer not in options or len(answer) < 5:
        return False

    # Options validation
    min_option_length = 2
    if (
        not 3 <= len(options) <= 5
        or len(set(options)) != len(options)
        or any(len(opt) < min_option_length for opt in options)
    ):
        return False

    # Question format check
    if question.startswith("What is") and len(question.split()) == 3:
        return False

    return True



"""Code quality improvement tools for the project."""

import logging
import os
import shlex
import subprocess  # nosec B404 - required for running formatters
import sys
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Sequence, Tuple

from pylint import lint

logger = logging.getLogger(__name__)

# Whitelist of allowed formatting commands with their exact arguments
ALLOWED_COMMANDS = {
    "autopep8": {
        "command": "autopep8",
        "required_args": ["--in-place", "--aggressive", "--aggressive"],
        "allow_files": True,
    },
    "isort": {
        "command": "isort",
        "required_args": ["--profile", "black"],
        "allow_files": True,
    },
    "black": {
        "command": "black",
        "required_args": ["--quiet"],
        "allow_files": True,
    },
}


@dataclass
class CommandResult:
    """Result of running a command."""

    success: bool
    output: str
    error: Optional[str] = None


class CodeFormatter:
    """Handles code formatting and quality checks."""

    def __init__(self, project_root: Optional[Path] = None):
        """Initialize with project root directory."""
        self.project_root = project_root or Path(__file__).parent.parent
        self.pylintrc = self.project_root / ".pylintrc"

    def is_safe_command(self, command_template: str) -> bool:
        """Validate if command is in the allowed whitelist.

        Security: This method ensures only pre-approved formatting commands are executed.
        Command templates must match exactly with allowed commands and their arguments.
        """
        parts = shlex.split(command_template)
        if not parts:
            return False

        cmd = parts[0]
        if cmd not in ALLOWED_COMMANDS:
            return False

        cmd_config = ALLOWED_COMMANDS[cmd]
        required_args = cmd_config["required_args"]

        # Check if all required arguments are present in order
        arg_pos = 1
        for req_arg in required_args:
            if arg_pos >= len(parts) or parts[arg_pos] != req_arg:
                return False
            arg_pos += 1

        return True

    def get_python_files(self, exclude_prefix: str = "") -> List[str]:
        """Get all Python files in the project except those starting with prefix."""
        python_files = []
        for root, _, files in os.walk(self.project_root):
            # Skip virtual environment and git directories
            if any(excluded in root for excluded in ["venv", ".git", "__pycache__"]):
                continue
            for file in files:
                if file.endswith(".py") and not file.startswith(exclude_prefix):
                    python_files.append(os.path.join(root, file))
        return python_files

    def run_command(self, command: Sequence[str], timeout: int = 30) -> CommandResult:
        """Run a shell command and return the result.

        Args:
            command: List of command arguments (first item is the command)
            timeout: Maximum time to wait for command completion

        Security:
        - Commands are validated against a strict whitelist
        - Arguments are properly quoted
        - shell=False prevents shell injection
        - Only formatting tools are allowed
        """
        if not command:
            return CommandResult(success=False, output="", error="Empty command")

        # Validate command is in whitelist
        cmd_str = " ".join(command)
        if not self.is_safe_command(cmd_str):
            return CommandResult(
                success=False, output="", error=f"Command not in whitelist: {cmd_str}"
            )

        try:
            # nosec B603 - Command is validated against strict whitelist and shell=False is used
            result = subprocess.run(
                command,
                shell=False,  # nosec B603
                check=True,
                capture_output=True,
                text=True,
                timeout=timeout,
            )
            return CommandResult(success=True, output=result.stdout, error=None)
        except subprocess.TimeoutExpired:
            return CommandResult(
                success=False, output="", error=f"Command timed out after {timeout} seconds"
            )
        except subprocess.CalledProcessError as e:
            return CommandResult(success=False, output=e.stdout or "", error=e.stderr)
        except Exception as e:
            return CommandResult(success=False, output="", error=str(e))

    def run_command_batch(
        self, command_template: str, files: List[str], batch_size: int = 10, timeout: int = 30
    ) -> CommandResult:
        """Run a command on batches of files in parallel."""
        if not files:
            return CommandResult(success=True, output="No files to process", error=None)

        # Split command template into base command and args
        cmd_parts = shlex.split(command_template)
        if not cmd_parts:
            return CommandResult(success=False, output="", error="Empty command template")

        cmd_name = cmd_parts[0]
        if cmd_name not in ALLOWED_COMMANDS:
            return CommandResult(
                success=False,
                output="",
                error=f"Command not in whitelist: {cmd_name}",
            )

        all_results = []
        total_batches = (len(files) + batch_size - 1) // batch_size

        with ThreadPoolExecutor() as executor:
            futures = []
            for i in range(0, len(files), batch_size):
                batch = files[i : i + batch_size]
                # Build command: base command + required args + files
                cmd = [cmd_name] + ALLOWED_COMMANDS[cmd_name]["required_args"] + batch
                futures.append(executor.submit(self.run_command, cmd, timeout))
                print(
                    f"\rProcessing batch {(i//batch_size)+1}/{total_batches}...", end="", flush=True
                )

            for future in futures:
                try:
                    result = future.result()
                    if not result.success:
                        return result
                    if result.output:
                        all_results.append(result.output)
                except Exception as e:
                    return CommandResult(success=False, output="", error=str(e))

        print()  # New line after progress
        return CommandResult(success=True, output="\n".join(all_results), error=None)

    def fix_code_style(self) -> bool:
        """Run automatic code fixers."""
        logger.info("Running automatic code fixes...")

        python_files = self.get_python_files("fix_code")
        if not python_files:
            logger.info("No Python files found to process!")
            return True

        logger.info("Found %d Python files to process", len(python_files))
        logger.info("Running autopep8...")

        result = self.run_command_batch("autopep8", python_files, batch_size=5, timeout=60)
        if not result.success:
            logger.error("Error running autopep8: %s", result.error)
            return False

        return True

    def format_code(self) -> bool:
        """Run code formatters."""
        logger.info("Running code formatters...")

        python_files = self.get_python_files("fix_code")

        # Run isort
        logger.info("Running isort...")
        result = self.run_command_batch("isort", python_files)
        if not result.success:
            logger.error("Error running isort: %s", result.error)
            return False

        # Run black
        logger.info("Running black...")
        result = self.run_command_batch("black", python_files)
        if not result.success:
            logger.error("Error running black: %s", result.error)
            return False

        logger.info("Code formatting complete!")
        return True

    def run_pylint(self) -> Tuple[bool, float]:
        """Run pylint and return success status and score."""
        python_files = self.get_python_files("fix_code")
        pylint_args = [f"--rcfile={self.pylintrc}", "--output-format=colorized"] + python_files

        try:
            run = lint.Run(pylint_args, exit=False)
            score = run.linter.stats.global_note
            return score >= 7.0, score
        except Exception as e:
            logger.error("Error running pylint: %s", e)
            return False, 0.0


def main():
    """Run code quality improvement pipeline."""
    logging.basicConfig(level=logging.INFO)
    logger.info("Starting code quality improvement pipeline...")

    formatter = CodeFormatter()

    if not formatter.fix_code_style():
        sys.exit(1)

    if not formatter.format_code():
        sys.exit(1)

    success, score = formatter.run_pylint()
    if success:
        logger.info("Pylint score: %.2f/10", score)
        if score >= 9.0:
            logger.info("Excellent code quality achieved!")
        else:
            logger.info("Acceptable code quality reached.")
    else:
        logger.error("Code quality check failed")
        sys.exit(1)

    logger.info("Code quality pipeline completed successfully!")


if __name__ == "__main__":
    main()
